module parallelization
!***********************************************************************
!
!     MPI processes management
!
!***********************************************************************

use iso_c_binding
implicit none

integer(c_size_t), parameter :: HOST_NAME_MAX = 255

type, public :: RankHostInfo
  integer :: rank, rankhostinfo
  character(HOST_NAME_MAX) :: hostname
end type RankHostInfo

type, public :: process

  integer, private :: root, comm, comm_index
  
contains

  procedure, pass :: init => new_process
  procedure, pass :: mpi_err_check => mpi_err_check

  procedure, pass :: isMaster => process_is_master
  procedure, pass :: getRoot => process_get_root
  procedure, pass :: getComm => process_get_comm
  procedure, pass :: getCommIndex => process_get_comm_index
  procedure, pass :: resetComm => process_reset_comm
  procedure, pass :: getRank => process_get_rank
  procedure, pass :: getProcessCount => process_get_process_count
  procedure, pass :: dedicate => process_dedicate
  procedure, pass :: states_range => process_states_range
  procedure, pass :: samples_range => process_samples_range

  procedure, pass :: gatherArrayIntracomm => process_gather_array_intracomm
  procedure, pass :: gatherArrayIntercomm => process_gather_array_intercomm

  procedure, pass :: gatherLArrayIntracomm => process_gather_Larray_intracomm

end type process

private :: HOST_NAME_MAX

contains


!======================================================================      


function process_getchar(this)

class(process) :: this
integer :: process_getchar

end function process_getchar


!======================================================================      


subroutine MPI_Err_Check(this, ierr, istatus)
implicit none
include 'mpif.h'

interface

function getchar() bind(C, name = "getchar")
use iso_c_binding
implicit none

integer(c_int) :: getchar

end function getchar

end interface

class(process) :: this
integer, intent(in) :: ierr
integer, intent(in), optional :: istatus(MPI_STATUS_SIZE)

character(1) :: freeze_on_error
integer :: length, ierr2, env_var_status
character(len = MPI_MAX_ERROR_STRING) :: errstr

if (ierr .ne. MPI_SUCCESS) then
  errstr = ' '
  call MPI_Error_string(ierr, errstr, length, ierr2)
  write(*, '(A I0 A A A)') 'MPI error code = ', ierr, ' has encountered: "', trim(errstr), '"'
  if (present(istatus)) then
!   The MPI_ERROR field of the status return is only set if the return from the MPI routine is MPI_ERR_IN_STATUS. 
    if (ierr .eq. MPI_ERR_IN_STATUS) then
      call MPI_Error_string(istatus(MPI_ERROR), errstr, length, ierr2)
      write(*, '(A I0 A A A)') 'MPI status code = ', ierr, ' has encountered: "', trim(errstr), '"'
    endif
  endif

  call get_environment_variable("FREEZE_ON_ERROR", freeze_on_error, length, env_var_status)
  if (env_var_status .lt. 1) then
    if ((length .ne. 0) .and. (freeze_on_error .ne. "0")) then
      write(*, '(A)') 'Press Enter to abort ...'
      length = getchar()
    endif
  endif
  stop
endif

end subroutine Mpi_Err_Check


!======================================================================      


!!
!! Setup MPI configuration: hostnames & groups.
!!
subroutine new_process(this)
implicit none
include 'mpif.h'

interface

function gethostname(hostname, length) bind(C, name = "gethostname")
use iso_c_binding
implicit none

integer(c_int) :: gethostname
character(kind = c_char), dimension(*), intent(in) :: hostname
integer(c_size_t), value, intent(in) :: length

end function gethostname

function getchar() bind(C, name = "getchar")
use iso_c_binding
implicit none

integer(c_int) :: getchar

end function getchar

end interface

class(process), intent(out) :: this
type(RankHostInfo) :: rhi
type(RankHostInfo), allocatable :: rhis(:)
integer, parameter :: root = 0
integer, allocatable :: hostranks(:)
integer :: mpi_ierr
integer :: rank, my_rank, nranks, ngroups, nhosts
character(HOST_NAME_MAX), allocatable :: groups(:)
character(HOST_NAME_MAX) :: hostname
logical :: seen
integer :: nhostranks
integer, allocatable :: group(:)
integer :: i, j, ireq
character :: ch
integer, allocatable :: requests(:)
integer :: worldGroup, hostComm, hostnameGroup
character(1) :: freeze_on_init
integer :: length, env_var_status
integer :: mpi_status(MPI_STATUS_SIZE)

this%root = 0
call this%resetComm()

! Figure out my host name.
if (gethostname(rhi%hostname, HOST_NAME_MAX) .ne. 0) then
  write(*, '(A, I0)') 'Error in gethostname'
  stop
endif

hostname = rhi%hostname

do i = 1, len(hostname)
  ch = hostname(i:i)
  if (ch .eq. c_null_char) then
    hostname(min(len(hostname), i + 1) : len(hostname)) = ' '
    exit
  endif
enddo

rhi%hostname = hostname

call MPI_Init(mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

! All MPI routines (except MPI_Wtime and MPI_Wtick) return an error value.
! Before the value is returned, the current MPI error handler is called.
! By default, this error handler aborts the MPI job. The error handler may be changed
! with MPI_Comm_set_errhandler (for communicators), MPI_File_set_errhandler (for files),
! and MPI_Win_set_errhandler (for RMA windows). The predefined error handler
! MPI_ERRORS_RETURN may be used to cause error values to be returned.
call MPI_Comm_set_errhandler(this%comm, MPI_ERRORS_RETURN, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

call MPI_Comm_rank(this%comm, rank, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)
call MPI_Comm_size(this%comm, nranks, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

if (this%isMaster()) then
  write(*, *) '------------------------------------'
  write(*, *) 'INITIALIZING MPI'
  write(*, *) '------------------------------------'
endif
	
! Collect host names from all nodes.
rhi%rank = rank
if (this%isMaster()) then
  allocate(rhis(0:nranks - 1))
endif
call MPI_Gather(rhi, sizeof(rhi), MPI_BYTE, rhis(0), &
  sizeof(rhi), MPI_BYTE, root, this%comm, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)
	
! Find groups on master only and send them to all other nodes.
! Gather nodes belonging to the same host name
! into communicators.
if (this%isMaster()) then
  ! Track already discovered groups of ranks with the same
  ! hostname.
  allocate(groups(0:nranks - 1))
  ngroups = 0
  do my_rank = 0, nranks - 1
    hostname = rhis(my_rank)%hostname
			
    ! Do not process again previously seen hostname.
    seen = .false.
    do i = 0, ngroups - 1
      if (hostname == groups(i)) then
        seen = .true.
        exit
      endif
    enddo

    if (seen) cycle

    groups(ngroups) = hostname
    ngroups = ngroups + 1

    ! Count how many ranks are there with the current hostname.
    nhostranks = 0
    do i = 0, size(rhis) - 1
      if (hostname == rhis(i)%hostname) then
        nhostranks = nhostranks + 1
      endif
    enddo

    ! Record those ranks.
    allocate(group(0:nhostranks - 1))
    nhostranks = 0
    do i = 0, size(rhis) - 1
      if (hostname .eq. rhis(i)%hostname) then
        group(nhostranks) = rhis(i)%rank
        nhostranks = nhostranks + 1
      endif
    enddo
      
    ! Send size of group and the ranks themselves to every
    ! member of group. Do not send to the master node himself.
    allocate(requests(0:nhostranks * 2 - 1))
    do i = 0, nhostranks - 1
      if (group(i) .eq. root) then
        ! Fill the hostranks on the master node w/o send/recv.
        allocate(hostranks(0:nhostranks - 1))
        hostranks(:) = group(:)
        cycle
      endif

      call MPI_Isend(nhostranks, 1, MPI_INT, group(i), 0, this%comm, requests(2 * i), mpi_ierr)
      call this%MPI_Err_Check(mpi_ierr)
      call MPI_Isend(group(0), nhostranks, MPI_INT, group(i), 1, this%comm, requests(2 * i + 1), mpi_ierr)
      call this%MPI_Err_Check(mpi_ierr)
    enddo
    do i = 0, 2 * nhostranks - 1
      if (my_rank == 2 * root) cycle
      if (my_rank == 2 * root + 1) cycle
      call MPI_Wait(requests(i), mpi_status, mpi_ierr)
      call this%MPI_Err_Check(mpi_ierr, mpi_status)
    enddo

    nhosts = ngroups
!    write(*, '(I0, A)') nhosts, ' host(s) participating:'

!    do i = 0, nhosts - 1
!      write(*, '(A)', advance = "no") groups(i)
!    enddo
!    write(*, *)
    flush(6)

    deallocate(group)
    deallocate(requests)
  enddo

  deallocate(groups)

else

  ! Receive my group size & ranks from master.
  nhostranks = 0
  call MPI_Recv(nhostranks, 1, MPI_INT, root, 0, this%comm, mpi_status, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr, mpi_status)
  allocate(hostranks(0:nhostranks - 1))
  call MPI_Recv(hostranks(0), nhostranks, MPI_INT, root, 1, this%comm, mpi_status, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr, mpi_status)

endif

! Tell everyone the number of available hosts.
! Can do MPI_Ibcast here, but it's not available in some implementations.
if (this%isMaster()) then
  allocate(requests(0:nranks - 2))
  ireq = 0
  do i = 0, nranks - 1
    if (i .eq. root) cycle
    call MPI_Isend(nhosts, 1, MPI_INT, i, 0, this%comm, requests(ireq), mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)
    ireq = ireq + 1
  enddo
else
  allocate(requests(0:0))
  call MPI_Irecv(nhosts, 1, MPI_INT, root, 0, this%comm, requests(0), mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)
endif
	
if (size(hostranks) .eq. 0) then
  write(*, '(A, A)') "The number of ranks in hostname group cannot be zero,", &
    " something went wrong, aborting"
  stop
endif
			
call MPI_Comm_group(this%comm, worldGroup, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

call MPI_Group_incl(worldGroup, size(hostranks), hostranks(0), hostnameGroup, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)
call MPI_Comm_create(this%comm, hostnameGroup, hostComm, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

do my_rank = 0, nranks - 1
  if (my_rank .eq. rank) then
!    write(*, '(A, I0, A, I0, A, A, A)', advance="no") &
!      'rank ', rank, ' (pid = ', getpid(), ') @ hostname "', trim(hostname), '"'
    if (size(hostranks) - 1 .ne. 0) then
!      write(*, '(A, I0, A)', advance="no") ', together with ', size(hostranks) - 1, ' other ranks:'
      do i = 0, size(hostranks) - 1
        if (rank .ne. hostranks(i)) then
!          write(*, '(A, I0)', advance="no") ' ', hostranks(i)
	    endif
      enddo
!      write(*, *)
    endif
  endif
  
  flush(6)
  
  call MPI_Barrier(this%comm, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)
enddo

if (this%isMaster()) write(*, *)

! Maybe wait on init, if we want to attach a debugger
! to some MPI processes.
call get_environment_variable("FREEZE_ON_INIT", freeze_on_init, length, env_var_status)
if (env_var_status .lt. 1) then
  if ((length .ne. 0) .and. (freeze_on_init .ne. "0")) then
    call MPI_Barrier(this%comm, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)

    if (this%isMaster()) then
      write(*, '(A)') 'Press Enter to continue ...'
      mpi_ierr = getchar()
    endif
  endif
endif

call MPI_Barrier(this%comm, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

deallocate(requests)

end subroutine new_process


!======================================================================      


function process_is_master(this, comm)
implicit none
include 'mpif.h'

class(process), intent(in) :: this
integer, intent(in), optional :: comm
logical :: process_is_master

integer :: mpi_ierr
integer :: rank

if (present(comm)) then
  call MPI_Comm_rank(comm, rank, mpi_ierr)
else
  call MPI_Comm_rank(this%comm, rank, mpi_ierr)
endif

call this%MPI_Err_Check(mpi_ierr)

process_is_master = .false.
if (rank .eq. 0) then
  process_is_master = .true.
endif

end function process_is_master


!======================================================================      


function process_get_root(this, comm)
implicit none

class(process), intent(in) :: this
integer, intent(in), optional :: comm
integer :: process_get_root

process_get_root = this%root

end function process_get_root


!======================================================================      


function process_get_comm(this)
implicit none

class(process), intent(in) :: this
integer :: process_get_comm

process_get_comm = this%comm

end function process_get_comm


!======================================================================


function process_get_comm_index(this)
implicit none

class(process), intent(in) :: this
integer :: process_get_comm_index

process_get_comm_index = this%comm_index

end function process_get_comm_index


!======================================================================


subroutine process_reset_comm(this)
implicit none
include 'mpif.h'

class(process), intent(out) :: this

this%comm = MPI_COMM_WORLD
this%comm_index = 0

end subroutine process_reset_comm


!======================================================================      


function process_get_rank(this, comm)
implicit none
include 'mpif.h'

class(process), intent(in) :: this
integer, intent(in), optional :: comm

integer :: process_get_rank
integer :: mpi_ierr

if (present(comm)) then
  call MPI_Comm_rank(comm, process_get_rank, mpi_ierr)
else
  call MPI_Comm_rank(this%comm, process_get_rank, mpi_ierr)
endif

call this%MPI_Err_Check(mpi_ierr)

end function process_get_rank


!======================================================================      


function process_get_process_count(this, comm)
implicit none
include 'mpif.h'

class(process), intent(in) :: this
integer, intent(in), optional :: comm
integer :: process_get_process_count

integer :: mpi_ierr

if (present(comm)) then
  call MPI_Comm_size(comm, process_get_process_count, mpi_ierr)
else
  call MPI_Comm_size(this%comm, process_get_process_count, mpi_ierr)
endif

call this%MPI_Err_Check(mpi_ierr)

end function process_get_process_count


!======================================================================      


!!!
!!! Dedicate process to communicator, based on the specified
!!! number of states and their sizes.
!!!
subroutine process_dedicate(this, nstates, sizes, parstates, initial)
implicit none
include 'mpif.h'

type :: RankCommInfo
  integer :: rank
  integer :: comm
end type RankCommInfo

interface

function lrint(x) bind(C, name = 'lrint')
use iso_c_binding
implicit none

integer(c_long) :: lrint
real(8), value, intent(in) :: x

end function lrint

end interface

class(process), intent(inout) :: this
integer, intent(in) :: nstates, sizes(nstates)
logical, intent(in) :: parstates, initial

integer :: i, mpi_ierr      
integer :: sizesTotalSum
integer, allocatable :: ranksPerState(:)
integer :: rank, ranksEvenDistPortion, ranksRemainder, remainingRanks, ranksForThisState
integer :: worldGroup, statesGroup
integer :: group_range(3)
integer :: state, state_count, statesComm, statesCommRootToWorld
integer :: statesCommRoot, statesCommRank, statesCommSize
integer :: mpiCommSplit

type(RankCommInfo) :: rci
type(RankCommInfo), allocatable :: rcis(:)

group_range = (/ 0, nstates - 1, 1 /)

! Calculate the size we have in all states in total.
! Then we will use this information to balance the # of processes
! assigned to per-state MPI communicators.
sizesTotalSum = 0
allocate(ranksPerState(nstates))
ranksPerState = 0

! First, reset process' communicator before new iteration.
call this%resetComm()

if (.not. parstates) then
  ! If parallel states mode is not activated,
  ! simply use plain WORLD communicator for all ranks.
  return
endif

if (initial) then

  ! We do not do any balancing on initial step.
  ! That is, ranks are distributed evenly.
  ranksEvenDistPortion = this%getProcessCount() / nstates
  
  if (ranksEvenDistPortion .lt. 1) then
    write(*, *) 'Not enough ranks to cover all per-state MPI communicators'
    stop
  endif
  
  do state_count = 1, nstates
    ranksPerState(state_count) = ranksEvenDistPortion
  enddo

  ! Distribute the remaining ranks as evenly as possible.
  ranksRemainder = mod(this%getProcessCount(), nstates)
  do state_count = 1, ranksRemainder
    ranksPerState(state_count) = ranksPerState(state_count) + 1
  enddo

else

  ! For efficiency, every process opens its own file, then nno values
  ! are gathered.
  call MPI_Comm_group(this%getComm(), worldGroup, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)
  call MPI_Group_range_incl(worldGroup, 1, group_range, statesGroup, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)
  call MPI_Comm_create(this%getComm(), statesGroup, statesComm, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)

  statesCommRoot = 0
  statesCommSize = 0
  if (statesComm .ne. MPI_COMM_NULL) then

    call MPI_Comm_size(statesComm, statesCommSize, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)
    if (statesCommSize .ne. nstates) then
      write(*, *) 'States communicator size must be equal to the number of states'
      write(*, *) 'something went wrong, aborting'
      stop
    endif

    statesCommRank = MPI_UNDEFINED
    call MPI_Comm_rank(statesComm, statesCommRank, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)

    state = statesCommRank + 1
    
    if (state .gt. nstates) then
      write(*, '(A, I0, A)') 'The state index ', state, ' is out of range'
    endif    
    if (sizes(state) .eq. 0) then
      write(*, *) 'The value of "sizes" cannot be zero'
      stop
    endif

    ! Gather sizes's into ranksPerState on master
    call MPI_Gather(sizes(state), 1, MPI_INT, ranksPerState, &
      1, MPI_INT, statesCommRoot, statesComm, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)
    
    if (statesCommRank == statesCommRoot) then

      ! Calculate the total sum of sizes's.
      sizesTotalSum = sum(ranksPerState)

      ! Assign each state a subset of ranks, based on the number of
      ! nno's in state.
      remainingRanks = this%getProcessCount()
      do state_count = 1, nstates - 1

        ranksForThisState = int(lrint(real(ranksPerState(state_count), 8) / &
          sizesTotalSum * this%getProcessCount()))
    
        ! Each state must be assigned at least one rank.
        if (ranksForThisState .eq. 0) ranksForThisState = ranksForThisState + 1
    
        ranksPerState(state_count) = ranksForThisState
        remainingRanks = remainingRanks - ranksForThisState
    
        if (remainingRanks .le. 0) then
          write(*, *) 'Not enough ranks to cover all per-state MPI communicators'
          stop
        endif

      enddo

      ranksPerState(nstates) = remainingRanks

    endif

    call MPI_Comm_free(statesComm, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)
    
  endif

  ! Translate statesComm's root to worldComm's rank.
  statesCommRootToWorld = MPI_UNDEFINED
  call MPI_Group_translate_ranks( &
    statesGroup, 1, statesCommRoot, worldGroup, statesCommRootToWorld, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)

  ! Broadcast the resulting ranksPerState to all nodes.
  call MPI_Bcast(ranksPerState, nstates, &
    MPI_INT, statesCommRootToWorld, this%getComm())
  call this%MPI_Err_Check(mpi_ierr)

  call MPI_Group_free(statesGroup, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)

  call MPI_Group_free(worldGroup, mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)

endif

if (sum(ranksPerState) .ne. this%getProcessCount()) then
  write(*, '(A I0 A I0 A)') 'Sum of ranks accross states (', sum(ranksPerState), &
    ')  must be equal to the number of processes (', this%getProcessCount(), ')'
  write(*, *) 'something went wrong, aborting'
  stop
endif

! Split initial communicator based on calculated # of ranks per communicator.
rci%rank = -1
rci%comm = -1

! Which communicator do I belong to ?
this%comm_index = 0
rank = this%getRank()
do state_count = 1, nstates
  
  if (rank .lt. ranksPerState(state_count)) then
    ! This # state is my communicator index.
    rci%rank = this%getRank()
    rci%comm = this%comm_index
    exit
  endif
    
  ! Advance to the next communicator.
  rank = rank - ranksPerState(state_count)
  this%comm_index = this%comm_index + 1

enddo

if ((rci%rank .eq. -1) .and. (rci%comm .eq. -1)) then
  write(*, '(A, I0)') 'Error assigning parallel state communicator index to rank', this%getRank()
  stop
endif

! Split the communicator based on ranks' communicator indexes.
call MPI_Comm_split(this%getComm(), this%comm_index, &
  this%getRank(), mpiCommSplit, mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

! Collect pairs (rank, this%comm_index) from all nodes.
if (this%isMaster()) then
  allocate(rcis(this%getProcessCount()))
else
  allocate(rcis(1))
endif

call MPI_Gather(rci, sizeof(rci), MPI_BYTE, rcis, &
    sizeof(rci), MPI_BYTE, this%getRoot(), this%getComm(), mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

flush(6)

call MPI_Barrier(this%getComm(), mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

if (this%isMaster()) then

  write(*, '(A)') 'Rank <-> state mapping :'

  do state_count = 1, nstates

    write(*, '(A, I0, A, I0, A)', advance = "no") 'state ', state_count, ' : ', ranksPerState(state_count), ' rank(s) : '

    do i = 1, size(rcis)
      if (rcis(i)%comm .eq. state_count - 1) then
        write(*, '(I0, A)', advance = "no") rcis(i)%rank, " "
      endif
    enddo

    write(*, *)

  enddo

  write(*, *)
  flush(6)

endif

call MPI_Barrier(this%getComm(), mpi_ierr)
call this%MPI_Err_Check(mpi_ierr)

this%comm = mpiCommSplit

deallocate(ranksPerState)

end subroutine process_dedicate


!======================================================================      


subroutine process_states_range(this, state_start, state_end)
implicit none
include 'mpif.h'

class(process), intent(in) :: this
integer, intent(inout) :: state_start, state_end

if (this%comm .ne. MPI_COMM_WORLD) then
  ! We are using parallel states
  state_start = this%comm_index + 1
  state_end = state_start
endif

end subroutine process_states_range


!======================================================================      


subroutine process_samples_range(this, sample_start, sample_end, state, rank)
implicit none
include 'mpif.h'

class(process), intent(in) :: this
integer, intent(inout) :: sample_start, sample_end
integer, intent(in) :: state
integer, intent(in), optional :: rank
integer :: this_getRank, noSamplesPerRank, samplesRemainder

if (present(rank)) then
  this_getRank = rank
else
  this_getRank = this%getRank()
endif

! Split between MPI ranks of either:
! - communicator corresponding to the respective parallel state
! - MPI_COMM_WORLD, if parallel states are not in use

noSamplesPerRank = state / this%getProcessCount()
samplesRemainder = mod(state, this%getProcessCount())
if (samplesRemainder .ne. 0) then
  if (this_getRank .lt. samplesRemainder) then
    sample_start = 1 + (noSamplesPerRank + 1) * this_getRank
    sample_end = sample_start + noSamplesPerRank
  else
    sample_start = 1 + (noSamplesPerRank + 1) * samplesRemainder + &
      (this_getRank - samplesRemainder) * noSamplesPerRank
    sample_end = sample_start + noSamplesPerRank - 1
  endif
else
  sample_start = 1 + noSamplesPerRank * this_getRank
  sample_end = sample_start + noSamplesPerRank - 1
endif

end subroutine process_samples_range


!======================================================================      


subroutine process_gather_array_intracomm(this, array, length)
implicit none
include 'mpif.h'

class(process), intent(in) :: this
real(8), intent(inout), dimension(:,:) :: array
integer, intent(in) :: length

integer :: mpi_ierr, rank
integer :: sample_start, sample_end
integer, allocatable :: requests(:)
integer :: mpi_status(MPI_STATUS_SIZE)

if (this%isMaster()) then
  allocate(requests(0:this%getProcessCount() - 1))
  requests = MPI_REQUEST_NULL

  do rank = 0, this%getProcessCount() - 1
    if (rank .eq. this%getRoot()) cycle
      
    call this%samples_range(sample_start, sample_end, length, rank)
    call MPI_Irecv(array(1, sample_start), &
      size(array, 1) * (sample_end - sample_start + 1) * c_sizeof(array(1, 1)), &
      MPI_BYTE, rank, rank, this%getComm(), requests(rank), mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)
  enddo
  do rank = 0, this%getProcessCount() - 1
    if (requests(rank) .eq. MPI_REQUEST_NULL) cycle

    call MPI_Wait(requests(rank), mpi_status, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr, mpi_status)
  enddo
    
  deallocate(requests)
else
  call this%samples_range(sample_start, sample_end, length, this%getRank())
  call MPI_Send(array(1, sample_start), &
    size(array, 1) * (sample_end - sample_start + 1) * c_sizeof(array(1, 1)), &
    MPI_BYTE, this%getRoot(), this%getRank(), this%getComm(), mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)
endif

end subroutine process_gather_array_intracomm


!======================================================================      


subroutine process_gather_array_intercomm(this, array_dst, array_src, nstates, state_count)

implicit none
include 'mpif.h'

class(process), intent(in) :: this
character, intent(out), dimension(:,:) :: array_dst
character, intent(in), dimension(:,:) :: array_src
integer, intent(in) :: nstates, state_count

integer :: mpi_ierr, state_count2
integer, allocatable :: requests(:)
integer :: mpi_status(MPI_STATUS_SIZE)

if (this%isMaster(MPI_COMM_WORLD)) then

  allocate(requests(nstates))
  requests = MPI_REQUEST_NULL

  do state_count2 = 1, nstates

!.. Do not receive from itself - copy explicitly.
    if (this%isMaster() .and. (state_count2 .eq. state_count)) then
      array_dst(1:size(array_dst, 1), state_count) = &
        array_src(1:size(array_src, 1), state_count)
      cycle
    endif

!.. Receive only from masters of each group.
!.. We do not provide explicit source ranks here (it's hard to compute them).
!.. Instead, we receive from any source, based on a tag identifiying
!.. each individual group.
    call MPI_Irecv(array_dst(1, state_count2), size(array_dst, 1), &
      MPI_BYTE, MPI_ANY_SOURCE, state_count2, MPI_COMM_WORLD, &
      requests(state_count2), mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)

  enddo

  do state_count2 = 1, nstates
!.. Do not receive from itself.
    if (requests(state_count2) .eq. MPI_REQUEST_NULL) cycle

    call MPI_Wait(requests(state_count2), mpi_status, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr, mpi_status)
  enddo

  deallocate(requests)

else

  ! Send only from master of each group. Do not send from itself.
  if (this%isMaster()) then
    call MPI_Send(array_src(1, state_count), size(array_src, 1), &
      MPI_BYTE, this%getRoot(MPI_COMM_WORLD), state_count, &
      MPI_COMM_WORLD, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)
  endif

endif


end subroutine process_gather_array_intercomm


!======================================================================      

subroutine process_gather_Larray_intracomm(this, array, length)
implicit none
include 'mpif.h'

class(process), intent(in) :: this
logical(C_BOOL), intent(inout), dimension(:,:) :: array
integer, intent(in) :: length

integer :: mpi_ierr, rank
integer :: sample_start, sample_end
integer, allocatable :: requests(:)
integer :: mpi_status(MPI_STATUS_SIZE)

if (this%isMaster()) then
  allocate(requests(0:this%getProcessCount() - 1))
  requests = MPI_REQUEST_NULL

  do rank = 0, this%getProcessCount() - 1
    if (rank .eq. this%getRoot()) cycle
      
    call this%samples_range(sample_start, sample_end, length, rank)
    call MPI_Irecv(array(1, sample_start), &
      size(array, 1) * (sample_end - sample_start + 1) * c_sizeof(array(1, 1)), &
      MPI_BYTE, rank, rank, this%getComm(), requests(rank), mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr)
  enddo
  do rank = 0, this%getProcessCount() - 1
    if (requests(rank) .eq. MPI_REQUEST_NULL) cycle

    call MPI_Wait(requests(rank), mpi_status, mpi_ierr)
    call this%MPI_Err_Check(mpi_ierr, mpi_status)
  enddo
    
  deallocate(requests)
else
  call this%samples_range(sample_start, sample_end, length, this%getRank())
  call MPI_Send(array(1, sample_start), &
    size(array, 1) * (sample_end - sample_start + 1) * c_sizeof(array(1, 1)), &
    MPI_BYTE, this%getRoot(), this%getRank(), this%getComm(), mpi_ierr)
  call this%MPI_Err_Check(mpi_ierr)
endif

end subroutine process_gather_Larray_intracomm

!======================================================================   

end module parallelization

