!***********************************************************************
!
!     Module parameter_module.f90: 
!     sets the parameters for the
!     model
!
!***********************************************************************

      module parameter_module

      use, intrinsic :: iso_c_binding

      use parallelization
      USE TasmanianSG
     
      implicit none
!      include 'mpif.h'

!.....things to set
      integer, parameter :: start_iter = 310    !where to start VFI
      integer, parameter :: end_iter   = 400  !where to end VFI
      integer, parameter :: min_level = 3  !minimum refinment level for SG
      integer, parameter :: max_level = 6  !maximum refinment level for SG
      integer, parameter :: prec = 8   !single: 4, double prec: 8      
      real(prec), parameter :: refinment_crit = 0.1  !criterion for refinment
      logical, parameter ::  policy_output = .true. !if false only the value function value will written except for final iteration  		                      
      logical, parameter :: binaryio = .false.   !0: no, 1: yes  for binary I/O      

!.....MPI switch
      logical, parameter :: parstates = .false. !turns on parallelization over states
      type(process) :: proc

!.....technical parameters 

      integer, parameter :: nrestarts = 30       !nglobal restarts
      integer, parameter :: debug = 0 !debug mode if 0 then no output; if >=5 then Ipopt has normal output
      integer, parameter :: seed_now = 12345 !random seed to generate observations for GPR
      real(prec), parameter :: bigbound = 1.0d+10 !value considered infinity by solver

!.....C interface
      integer(C_INT), target :: debug_target = debug
      integer(C_INT), target :: start_iter_target = start_iter 	 
      TYPE(C_PTR), BIND(C, NAME='debug_ptr') :: debug_ptr   
      TYPE(C_PTR), BIND(C, NAME='start_iter_ptr') :: start_iter_ptr   
!.....Solver choice
      integer, parameter :: which_optimizer = 2 ! 1: IPOPT ; 2: SNOPT
!.....C interface
      integer(C_INT), target :: which_optimizer_target = which_optimizer
      TYPE(C_PTR), BIND(C, NAME='which_optimizer_ptr') :: which_optimizer_ptr  

 
!.....Econ model parameters 
      integer(C_INT), parameter :: dim_model = 2 	!dimension in which we compute the bgmm
      integer, parameter :: dim_sim = dim_model        !dimension of continuous state space
      integer, parameter :: n_disc_state = 2          !no discrete states
      integer, parameter :: no_integnodes = n_disc_state     !no. of integrationnodes; is equal to disc states here since we have no cont. shocks
      real(prec), parameter :: beta = 0.98
      real(prec), parameter :: sigma = 0.5
      real(prec), parameter :: cmin  = 0.0001   
      real(prec), parameter :: psi  = 2.0                   
      real(prec), parameter :: util_term = 1.0

      real(C_DOUBLE), save, dimension(1:no_integnodes) :: outcomes = (/ &
      & 0.,2./)     !payoffs

!.....C interface
      integer(C_INT), target :: dim_model_target = dim_model 	 
      integer(C_INT), target :: n_disc_state_target = n_disc_state     
      TYPE(C_PTR), BIND(C, NAME='dim_model_ptr') :: dim_model_ptr   
      TYPE(C_PTR), BIND(C, NAME='n_disc_state_ptr') :: n_disc_state_ptr   

!.....number of unknowns in constraints       
      integer, parameter :: nineq = 31              !no. of inequalities
      integer, parameter :: neq   = 16              !no. of equalities
      integer, parameter :: nconstr = neq + nineq   !no. of constraints
      integer, parameter :: nvars = 30              !no. of variables
      integer, parameter :: no_nz_jac = 127        !no. of nonzero entries in Jac
      integer, parameter :: no_nz_obj = 6           !no. of nonzero entries in objective for problem
     
!.....location for restart files
      character(256) :: restart_dir = 'restart/'
      logical, save :: restart_on = .true.

      
!.....penalty parameters 
      real(prec), parameter :: comppara = 0.            !complementarity relaxation parameter
      real(prec), parameter :: epsmat = 0.00001          !parameter to relax the sdp condition


!.....state space bounds
      real(C_DOUBLE),  dimension(dim_model*n_disc_state) :: state_up_vec
      real(C_DOUBLE),  dimension(dim_model*n_disc_state) :: state_down_vec
      real(C_DOUBLE), save, target, dimension(1:dim_model) :: statespace_up   = (/ &
      & 70.2124, 70.2124 /)
      real(C_DOUBLE), save, target, dimension(1:dim_model) :: statespace_down = (/ &
      &15.3193, 15.3193/)    
  
!.....tasmanian grids
      type(TasmanianSparseGrid) :: grid_new(n_disc_state),grid_old(n_disc_state)   !grid objects
       
!---------------------------------------------------------------------- 

     real(prec), parameter :: Pi = 3.141592653589793     
     real(prec), parameter :: E = 2.718281828459045

!.......................................................................................

 
      contains

!.......................................................................................

!******************************************************************************************
!----------Define pointers so that we can access vars in C
!******************************************************************************************

  SUBROUTINE init_paramodule
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_LOC

    dim_model_ptr    = C_LOC(dim_model_target)
    n_disc_state_ptr = C_LOC(n_disc_state_target)
    debug_ptr = C_LOC(debug_target)
    which_optimizer_ptr = C_LOC(which_optimizer_target)
    start_iter_ptr = C_LOC(start_iter_target)

  END SUBROUTINE init_paramodule

!******************************************************************************************
!----------Enable access to parallelization module in C
!******************************************************************************************

  SUBROUTINE init_process
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_LOC

    call new_process(proc)   

  END SUBROUTINE init_process

  SUBROUTINE is_master(ismaster)
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_LOC

    logical(c_bool), intent(out) :: ismaster

    ismaster = proc%isMaster()   

  END SUBROUTINE is_master

!*********************************************************************************************
!-----------Transfer stuff over from C++ to be acessible in FORTRAN
!*********************************************************************************************

!  SUBROUTINE transfer_paramodule(handle_BGM_tmp,handle_BGM_inf_tmp,halton_class_tmp)
!    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_LOC

!      integer(8), intent(in)     :: handle_BGM_tmp,handle_BGM_inf_tmp
!      integer(8), intent(in)     :: halton_class_tmp
!      integer :: indx

!      handle_BGM = handle_BGM_tmp
!      handle_infeas_BGM = handle_BGM_inf_tmp
!      halton_class = halton_class_tmp
!!      write(*,*) "Upper bounds"
!!      write(*,*) state_up_vec(:)
!!      write(*,*) "Lower bounds"
!!      write(*,*) state_down_vec(:)


!  END SUBROUTINE transfer_paramodule



!******************************************************************************************
!----------Quick Sort for an array
!******************************************************************************************

!.....Sorts matrix A with respect to the last row entry in the matrix
      
      recursive subroutine QsortC(A)
        real(prec), intent(in out), dimension(:,:) :: A
        integer :: iq
        integer, dimension(2) :: tmp_shape

        tmp_shape = shape(A)
      
        if(tmp_shape(1) > 1) then
           call Partition(A, iq)
           call QsortC(A(:iq-1,:))
           call QsortC(A(iq:,:))
        endif
      end subroutine QsortC
            
      subroutine Partition(A, marker)
        real(prec), intent(in out), dimension(:,:) :: A
        integer, intent(out) :: marker
        integer :: i, j
        real(prec), allocatable :: temp(:)
        real(prec) :: x      ! pivot point
        integer, dimension(2) :: tmp_shape

        tmp_shape = shape(A)
      
        allocate(temp(tmp_shape(2)))

        x = A(1,tmp_shape(2))
        i= 0
        j= tmp_shape(1) + 1
      
        do
           j = j-1
           do
              if (A(j,tmp_shape(2)) <= x) exit
              j = j-1
           end do
           i = i+1
           do
              if (A(i,tmp_shape(2)) >= x) exit
              i = i+1
           end do
           if (i < j) then
              ! exchange A(i) and A(j)
              temp(:) = A(i,:)
              A(i,:) = A(j,:)
              A(j,:) = temp(:)
           elseif (i == j) then
              marker = i+1
              return
           else
              marker = i
              return
           endif
        end do

        deallocate(temp)
      
      end subroutine Partition

!******************************************************************************************


      
!======================================================================  

      end module parameter_module

!***********************************************************************
      
